#!/usr/bin/python
# -*- coding: utf-8 -*-

# 
# MeCaBrola
# 
# The MIT License (MIT)
#
# Copyright (c) 2012- Takashi Sakamoto
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
# associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions
# of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import sys
# MeCabのPythonモジュール
import MeCab
# Mbrolaのphoファイルを一時的に書きだすため
import tempfile
# Mbrola及びaplayのプロセスを実行するため
import subprocess

# 音声の基本ピッチと振れ幅 (Hz)
tone = 250
tone_range = 30

# 1モーラの時間
mora_duration = 150

#  mbrola pho database
#   http://www.tcts.fpms.ac.be/synthesis/mbrola/mbrcopybin.html
# ./jp3/jp3にすると女声になるけど、基本ピッチを高くしないとちょっと怖い
dbpath = "./jp1/jp1"

# Mbrola用変換テーブル
yomi = {
#   母音
u"ア": "a",
u"イ": "i",
u"ウ": "u",
u"エ": "e",
u"オ": "o",
#  母音と子音
u"カ": "k,a",
u"キ": "k,i",
u"ク": "k,u",
u"ケ": "k,e",
u"コ": "k,o",
u"サ": "s,a",
u"シ": "S,i",
u"ス": "s,u",
u"セ": "s,e",
u"ソ": "s,o",
u"タ": "t,a",
u"チ": "tS,i",
u"ツ": "ts,u",
u"テ": "t,e",
u"ト": "t,o",
u"ナ": "n,a",
u"ニ": "n,i",
u"ヌ": "n,u",
u"ネ": "n,e",
u"ノ": "n,o",
u"ハ": "h,a",
u"ヒ": "h,i",
u"フ": "h,u",
u"ヘ": "h,e",
u"ホ": "h,o",
u"マ": "m,a",
u"ミ": "m,i",
u"ム": "m,u",
u"メ": "m,e",
u"モ": "m,o",
u"ヤ": "j,a",
u"ユ": "j,u",
u"ヨ": "j,o",
u"ラ": "rr,a",
u"リ": "rr,i",
u"ル": "rr,u",
u"レ": "rr,e",
u"ロ": "rr,o",
u"ワ": "w,a",
u"ヲ": "o",
u"ン": "N",
# 濁音
u"ガ": "g,a",
u"ギ": "g,i",
u"グ": "g,u",
u"ゲ": "g,e",
u"ゴ": "g,o",
u"ザ": "z,a",
u"ジ": "dZ,i",
u"ズ": "z,u",
u"ゼ": "z,e",
u"ゾ": "z,o",
u"ダ": "d,a",
u"ヂ": "dZ,i",
u"ヅ": "z,u",
u"デ": "d,e",
u"ド": "d,o",
u"バ": "b,a",
u"ビ": "b,i",
u"ブ": "b,u",
u"ベ": "b,e",
#  半濁音
u"ボ": "b,o",
u"パ": "p,a",
u"ピ": "p,i",
u"プ": "p,u",
u"ペ": "p,e",
u"ポ": "p,o"
}

# MeCabがNAIST Japanese Directionaryから取得する形態素情報
#   0.表層形
#   1.品詞
#   2.品詞細分類1
#   3.品詞細分類2
#   4.品詞細分類3
#   5.活用形
#   6.活用型
#   7.原形
#   8.読み
#   9.発音
def tokenizing_text(string):
	tokens = []
	tokenizer = MeCab.Tagger("-O wakati")
	result = tokenizer.parseToNode(string)
	# skip BOS
	result = result.next
	while result:
		if result.surface != "":
			elements = result.feature.split(",")
			
			token = [0] * 6
			if len(elements) < 8:
				token[0] = result.surface
			else:
				token[0] = elements[8]
			token[1] = result.surface
			token[2] = elements[0]
			token[3] = elements[1]
			token[4] = elements[2]
			token[5] = elements[3]
			
			# ここまで、stringはただのビット列だったのでそのまま扱うと1バイト処理されてしまう。
			# そこでpython unicodeオブジェクトに変換する。
			# naist-jdicからutf-8で符号化された文字列が取得できることが前提なので
			# MeCabのデフォルト辞書を確認してください。
			for (i,element) in enumerate(token):
				token[i] = unicode(element, "utf-8")
			
			tokens.append(token)
		result = result.next
	return tokens

# NAIST Japanese Dictionary version 0.4.0 ユーザーズマニュアル
# http://sourceforge.jp/projects/naist-jdic/docs/NAIST-JDIC-ja.pdf/ja/1/NAIST-JDIC-ja.pdf.pdf
#  5. IPA 品詞体系
#  5.1 名詞
#  5.2 接頭詞
#  5.3 動詞
#  5.4 形容詞
#  5.5 副詞
#  5.6 連体詞
#  5.7 接続詞
#  5.8 助詞
#  5.9 助動詞
#  5.10 感動詞
#  5.11 記号
#  5.12 フィラー
#  5.13 その他
def generating_lines(morphemes):
	temps = []
	candidates = []
	lines = []
	
	# 配列の前処理
	def pre_convert():
		def temps_append(symbol, index, letter, morpheme):
			temp = [symbol, index, letter, morpheme[1], morpheme[2], morpheme[3], morpheme[4]]
			temps.append(temp)
			return
		
		for morpheme in morphemes:
			for (i,letter) in enumerate(morpheme[0]):
				# 促音文字はこの時点では仮の記号とする
				if letter == u"ッ":
					temps_append("q", i, letter, morpheme)
				# 拗音文字はこの時点では仮の記号とする
				elif letter == u"ャ":
					temps_append("xj", i, letter, morpheme)
					temps_append("a", i, letter, morpheme)
				elif letter == u"ュ":
					temps_append("xj", i, letter, morpheme)
					temps_append("u", i, letter, morpheme)
				elif letter == u"ョ":
					temps_append("xj", i, letter, morpheme)
					temps_append("o", i, letter, morpheme)
				# 長音文字はこの時点では仮の記号とする
				elif letter == u"ー":
					temps_append("-", i, letter, morpheme)
				# 小さなア/イ/ウ/エ/オはこの時点では仮の記号とする
				# 「シャバドゥビタッチ変身」を喋らせてみたかっただけなんだ、反省はしている。
				elif letter == u"ァ":
					temps_append("xa", i, letter, morpheme)
				elif letter == u"ィ":
					temps_append("xi", i, letter, morpheme)
				elif letter == u"ゥ":
					temps_append("xu", i, letter, morpheme)
				elif letter == u"ェ":
					temps_append("xe", i, letter, morpheme)
				elif letter == u"ォ":
					temps_append("xo", i, letter, morpheme)
				# 上記以外の文字は上のディクショナリから引けるはず
				elif letter in yomi:
					elements = yomi[letter].split(",")
					# 一文字は1つあるいは2つの記号から成る
					if len(elements) < 2:
						temps_append(elements[0], i, letter, morpheme)
					else:
						temps_append(elements[0], i, letter, morpheme)
						temps_append(elements[1], i, letter, morpheme)
				# 引けなかったら無音記号にする
				else:
					temps_append("_", i, "", morpheme)
				continue
			continue
		return
	
	# 配列の本処理
	def convert():
		for (i,temp) in enumerate(temps):
			# 仮の促音記号は次の記号に影響する
			if temp[0] == "q":
				temps[i+1][0] = "Q" + temps[i+1][0]
				temps[i][0] = ""
			# サ行とザ行とチ行に続く拗音記号は2つの記号で表現する（促音記号も考慮する）
			# それ以外の拗音記号は3つの記号で表現する
			elif temp[0] == "xj":
				if temps[i-2][0] in ["QS", "QdZ", "QtS", "S", "dZ", "tS"]:
					temps[i-1][0] = ""
					temps[i][0] = ""
				else:
					temps[i-1][0] = ""
					temps[i][0] = "j"
				temps[i][1] = temps[i-2][1]
				temps[i][2] = temps[i-2][2]
				temps[i+1][1] = temps[i-2][1]
				temps[i+1][2] = temps[i-2][2]
			# 小さなア/イ/ウ/エ/オは前の記号に含める
			elif temp[0][0] == "x":
				temps[i][0] = temp[0][1]
				temps[i][1] = temps[i-1][1]
				temps[i][2] = temps[i-1][2]
			# 長音は前の記号を繰り返すことにする
			elif temp[0] == "-":
				temps[i][0] = temps[i-1][0]
			continue
		return
	
	# 配列の後処理
	def post_convert():
		for (i,temp) in enumerate(temps):
			# 記号のあるものだけ残す
			if temp[0] != "":
				candidates.append(temp)
			continue
		return
	
	# 制約の適用。促音を中心に。
	def apply_restriction():
		def append_sokuon_alternative(candidate):
			lines.append(["_", 100+candidate[1], u"促音の代理", candidate[3], candidate[4], candidate[5], candidate[6]])
			lines.append(["_", 100+candidate[1], u"促音の代理", candidate[3], candidate[4], candidate[5], candidate[6]])
			return
		
		def append_candidate(sign, candidate):
			lines.append([sign, candidate[1], candidate[2], candidate[3], candidate[4], candidate[5], candidate[6]])
			return
		
		for (i, candidate) in enumerate(candidates):
			# 最後の要素でなかったら評価。何故なら次の要素も合わせて評価できるように。
			if candidate != candidates[-1]:
				# 促音に関する制限を適用
				# jp1/jp3は、*-Q/Q-*に対応する収録データが多数不足している
				# 詳しく知りたい人はデータベースのバイナリ読むときっと幸せになれます
				# 大まかに言うと、バイナリはヘッダ部、エントリー部、PCMデータ部からなり、
				# 1. 0x1Cバイト目からエントリー部が始まる
				# 2. エントリーの構造は「0x00 ASCIIの1〜2バイト 0x00 ASCIIが1〜2バイト 0x00 5バイト」
				#    例: 「0x 72 72 00 6F 00 BA 01 1A 1A」これは「rr-o（7272と6F）」のエントリーを表す
				# 3. エントリーのASCIIを使って表を作成すると、合成可能な記号の組み合わせ表となる
				if candidate[0] == "QS" and candidates[i-1][0] in ["o"]:
					append_sokuon_alternative(candidate)
					append_candidate("S", candidate)
				elif candidate[0] == "Qs" and (candidates[i-1][0] in ["o"] or candidates[i+1][0] in ["u"]):
					append_sokuon_alternative(candidate)
					append_candidate("s", candidate)
				elif candidate[0] == "Qd" and (candidates[i-1][0] in ["u", "o"] or candidates[i+1][0] in ["a", "e"]):
					append_sokuon_alternative(candidate)
					append_candidate("d", candidate)
				elif candidate[0] == "Qts" and candidates[i-1][0] in ["a", "i", "e", "o"]:
					append_sokuon_alternative(candidate)
					append_candidate("ts", candidate)
				elif candidate[0] == "QtS":
					append_sokuon_alternative(candidate)
					append_candidate("tS", candidate)
				elif candidate[0] == "Qz":
					append_sokuon_alternative(candidate)
					append_candidate("z", candidate)
				elif candidate[0] == "QdZ":
					append_sokuon_alternative(candidate)
					append_candidate("dZ", candidate)
				elif candidate[0] == "Qp" and candidates[i+1][0] == "j":
					append_sokuon_alternative(candidate)
					append_candidate("p", candidate)
				else:
					append_candidate(candidate[0], candidate)
			else:
				append_candidate(candidate[0], candidate)
			continue
	
	# 配列の調整。継続時間とピッチを付与する
	# Mbrolaはピッチペアを複数与えることで単音にイントネーションを与えることができるが、今回はそれを使ってない
	# 中国語の音声合成はきっと大変だろうなぁ。。。
	def tuning():
		count = 0
		type = u""
		for (i,line) in enumerate(lines):
			# 配列の末尾に注意しつつ、1モーラを構成するものをグループにする
			if i+1 != len(lines) and line[1] == lines[i+1][1] and line[2] == lines[i+1][2]:
				count += 1
			else:
				# 前の品詞と異なったらピッチ下げる
				if line[3] != type:
					base_tone = tone - tone_range / 2
					type = line[3]
				elif i+1 != len(lines):
					# 次の品詞が異なるならピッチ結構下げる
					if  line[3] != lines[i+1][3]:
						base_tone = tone - tone_range
					# 次が記号ならピッチちょっと下げる（EOSを想定）
					elif lines[i+1][0] == "_":
						base_tone = tone - tone_range / 2
					type = line[3]
				else:
					base_tone = tone
				
				# 品詞からピッチを決定
				if line[3] in [u"動詞", u"名詞", u"形容詞", u"連体詞"]:
					line_tone = base_tone + tone_range
				else:
					line_tone = base_tone - tone_range
				
				# 1モーラの記号数から単音一つあたりの継続時間を決定
				line_duration = int(mora_duration / (count + 1) )
				
				# lineの作成と置き換え
				j = i - count
				while j <= i:
					# 促音記号があったら２倍の継続時間とする
					if lines[j][0][0] == "Q":
						lines[j] = [lines[j][0], line_duration * 2, 0, line_tone]
					else:
						lines[j] = [lines[j][0], line_duration, 0, line_tone]
					j += 1
				count = 0
			continue
		return

	# メイン処理
	pre_convert()
	convert()
	post_convert()
	apply_restriction()
	tuning()
	
	return lines

# ファイル書き出し
def write_pho_file(handle, line):
	for line in line:
		line = "\t".join(map(str, line)) + "\n"
		handle.write(line)
	
	handle.write("_\t90\t\n")
	handle.flush()
	return

# サブプロセスの起動
def run_subprocess(handle, dbpath):
	# もし何らかのコマンドライン引数を伴って実行されていたら、ファイル内容をダンプ
	# ロングオプション使いたかったけど面倒そうだったのでパス
	if ( len(sys.argv) > 1 ):
		command = "cat " + handle.name
		output = subprocess.check_output(command, shell=True)
		print output
	
	# 喋らせる
	command = "mbrola {0} {1} -.au | aplay".format(dbpath, handle.name)
	output = subprocess.check_output(command, shell=True)
	return output

# メイン部分
#  発言を標準入力。入力があるまでブロッキング
string = raw_input("何を喋りたい？: ")
#  この時点ではまだpython unicode stringになってない
tokens = tokenizing_text(string)
lines = generating_lines(tokens)
handle = tempfile.NamedTemporaryFile("w")
write_pho_file(handle, lines)
log = run_subprocess(handle, dbpath)
handle.close()
print log


